module.exports = (function () {
    var type = require("type-detect");
    var isNil = require("is-nil");
    var hasValue = function (x) { return !isNil(x); };
    var clone = require("clone");
    var associative = require("associative");
    var ArrayX = (function () {
        function ArrayX() {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            var Arr = new Array();
            args.forEach(function (arg) {
                if (console && console.warn && type(arg).toLocaleLowerCase() === "array")
                    console.warn("Array arg \"" + arg + "\" passed to ArraX constructor");
                Arr.push(arg);
            });
            Arr["toArray"] = function () {
                var arr = new Array(this.length);
                var self = this;
                for (var i = 0; i < self.length; i++)
                    arr[i] = self[i];
                return arr;
            };
            Arr["distinct"] = function () {
                var arr = [];
                var self = this;
                for (var i = 0; i < self.length; i++) {
                    if (arr.indexOf(self[i]) < 0)
                        arr.push(self[i]);
                }
                return new (ArrayX.bind.apply(ArrayX, [void 0].concat(arr)))();
            };
            Arr["union"] = function (array) {
                var arr = [];
                var self = this;
                for (var i = 0; i < self.length; i++)
                    arr.push(self[i]);
                if (type(array) === "Array")
                    for (var k = 0; k < array.length; k++)
                        arr.push(array[k]);
                return new (ArrayX.bind.apply(ArrayX, [void 0].concat(arr)))();
            };
            Arr["accumulate"] = function (initial, accum) {
                var self = this;
                var val = initial; // initial parameter is required
                for (var i = 0; i < self.length; i++)
                    val = accum ? accum(val, self[i]) : null;
                return val;
            };
            Arr["select"] = function (func) {
                var arr = [];
                var self = this;
                for (var i = 0; i < self.length; i++)
                    arr.push(func ? func(self[i], i) : self[i]);
                return new (ArrayX.bind.apply(ArrayX, [void 0].concat(arr)))();
            };
            Arr["where"] = function (func) {
                var arr = [];
                var self = this;
                for (var i = 0; i < self.length; i++)
                    if (isNil(func) || func(self[i], i) === true)
                        arr.push(self[i]);
                return new (ArrayX.bind.apply(ArrayX, [void 0].concat(arr)))();
            };
            /**
             * Sorts an array in ascending order.
             * @param func The function that generates the primitive values used to actually sort this array
             */
            Arr["orderBy"] = function (func) {
                var arr = clone(this).sort(function (a, b) {
                    var objA = func ? func(a) : a;
                    var objB = func ? func(b) : b;
                    var comparison = null;
                    if (objA && objB) {
                        if (type(objA) === "number")
                            comparison = objA - objB;
                        else if (type(objA) === "date")
                            comparison = objA.getTime() - objB.getTime();
                        else
                            comparison = objA.toString().localeCompare(objB.toString());
                    }
                    else {
                        if (!objA && !objB)
                            comparison = 0;
                        else if (!objA)
                            comparison = -1;
                        else
                            comparison = 1;
                    }
                    return comparison;
                });
                return new (ArrayX.bind.apply(ArrayX, [void 0].concat(arr)))();
            };
            /**
             * Sorts an array in descending order.
             * @param func The function that generate the primitive values used to actually sort this array
             */
            Arr["orderByDesc"] = function (func) {
                var self = this;
                var ordered = self.orderBy(func);
                return new (ArrayX.bind.apply(ArrayX, [void 0].concat(ordered.reverse())))();
            };
            /**
             * Returns an new array containing this array's elements in reverse order.
             */
            Arr["invert"] = function () {
                var arr = Arr["toArray"]();
                return arr.reverse();
            };
            Arr["firstOrDefault"] = function (func) {
                var self = this;
                if (isNil(func))
                    return self.length > 0 ? self[0] : null;
                for (var i = 0; i < self.length; i++)
                    if (func(self[i]) === true)
                        return self[i];
                return null;
            };
            Arr["first"] = function (func) {
                var self = this;
                var firstOrDefault = self.firstOrDefault(func);
                if (isNil(firstOrDefault))
                    throw new Error("First item not found");
                else
                    return firstOrDefault;
            };
            Arr["lastOrDefault"] = function (func) {
                var self = this;
                if (isNil(func))
                    return self.length > 0 ? self[self.length - 1] : null;
                for (var i = (self.length - 1); i >= 0; i--)
                    if (func(self[i]) === true)
                        return self[i];
                return null;
            };
            Arr["last"] = function (func) {
                var self = this;
                var lastOrDefault = self.lastOrDefault(func);
                if (isNil(lastOrDefault))
                    throw new Error("Last item not found");
                else
                    return lastOrDefault;
            };
            Arr["all"] = function (func) {
                var self = this;
                return self.where(func).length === this.length;
            };
            Arr["any"] = function (func) {
                var self = this;
                return self.firstOrDefault(func) !== null;
            };
            Arr["take"] = function (count) {
                var arr = [];
                var self = this;
                for (var i = 0; (i < count && i < self.length); i++)
                    arr.push(self[i]);
                return new (ArrayX.bind.apply(ArrayX, [void 0].concat(arr)))();
            };
            Arr["skip"] = function (count) {
                var arr = [];
                var self = this;
                for (var i = count; i < self.length; i++)
                    arr.push(self[i]);
                return new (ArrayX.bind.apply(ArrayX, [void 0].concat(arr)))();
            };
            Arr["sum"] = function (func) {
                var sum = 0;
                var self = this;
                for (var i = 0; i < self.length; i++)
                    sum += (func ? func(self[i]) : 0);
                return sum;
            };
            Arr["avg"] = function (func) {
                var self = this;
                var count = this.count(func);
                if (count === 0)
                    return 0;
                else
                    return self.sum(func) / count;
            };
            Arr["count"] = function (func) {
                var count = 0;
                var self = this;
                for (var i = 0; i < self.length; i++)
                    if (true == (func ? func(self[i]) : self[i]))
                        count++;
                return count;
            };
            Arr["min"] = function (func) {
                var min = Number.POSITIVE_INFINITY;
                var self = this;
                for (var i = 0; i < self.length; i++) {
                    var value = (func ? func(self[i]) : self[i]);
                    if (value < min)
                        min = value;
                }
                return min;
            };
            Arr["max"] = function (func) {
                var max = Number.NEGATIVE_INFINITY;
                var self = this;
                for (var i = 0; i < self.length; i++) {
                    var value = (func ? func(self[i]) : self[i]);
                    if (value > max)
                        max = value;
                }
                return max;
            };
            /*
             * Groups the array values into a dictionary of arrays.
             * Grouping keys are determined by applying the <func> function arg to each element
             */
            Arr["groupBy"] = function (func) {
                var self = this;
                var groups = {};
                for (var i = 0; i < self.length; i++) {
                    var key = func(self[i]);
                    if (!groups.hasOwnProperty(key))
                        groups[key] = new X.Array();
                    groups[key].append(self[i]);
                }
                return groups;
            };
            /*
             * Groups this array into an dictionary of <keyCount> arrays,
             * each containing the same number of elements, except possibly the last array,
             * which may have less elements than the others
             * if the length of this array is not divisible by <keyCount>
             */
            Arr["groupByNumber"] = function (keyCount) {
                var groups = {};
                var groupLength = Math.floor(this.length / keyCount);
                var remainder = this.length % this.count();
                for (var i = 0; i < keyCount; i++) {
                    if (!groups.hasOwnProperty(i))
                        groups[i] = new X.Array();
                    groups[i].append(clone(this).slice(i * groupLength, (i + 1) * groupLength));
                }
                return groups;
            };
            /*
             * groups this array according to <key> generator function and then
             * aggregates the resulting dictionary over <value> generator function
             */
            Arr["groupAggregate"] = function (key, value, aggregation) {
                // if key is passed as generator function, use it, 
                // else create generator function that assumes key is a property name. 
                // If key is null/missing, the string representation of an object is used as the key. 
                var keyGenerator = type(key) === "function"
                    ? key
                    : function (datum) { return key
                        ? datum[key.toString()]
                        : datum.toString(); };
                //Same for value
                var valueGenerator = type(value) === "function"
                    ? value
                    : function (datum) { return value
                        ? datum[value.toString()]
                        : datum.toString(); };
                var groups = this.groupBy(keyGenerator);
                Object.keys(groups)
                    .forEach(function (key) {
                    groups[key] = groups[key][aggregation](valueGenerator);
                });
                return groups;
            };
            Arr["removeItem"] = function (item) {
                var arr = clone(this);
                var _index = arr.indexOf(item);
                while (_index >= 0) {
                    arr.splice(_index, 1);
                    _index = arr.indexOf(item);
                }
                return new (ArrayX.bind.apply(ArrayX, [void 0].concat(arr)))();
            };
            Arr["removeAt"] = function (index, count) {
                var arr = clone(this);
                arr.splice(index, count);
                return new (ArrayX.bind.apply(ArrayX, [void 0].concat(arr)))();
            };
            Arr["insert"] = function (index) {
                var items = [];
                for (var _i = 1; _i < arguments.length; _i++) {
                    items[_i - 1] = arguments[_i];
                }
                var arr = clone(this);
                arr.splice.apply(arr, [index, 0].concat(items));
                return new (ArrayX.bind.apply(ArrayX, [void 0].concat(arr)))();
            };
            Arr["append"] = function () {
                var items = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    items[_i] = arguments[_i];
                }
                var arr = clone(this);
                arr.splice.apply(arr, [this.length, 0].concat(items));
                return new (ArrayX.bind.apply(ArrayX, [void 0].concat(arr)))();
            };
            Arr["flatten"] = function () {
                var mutable = false;
                var self = this;
                var toString = Object.prototype.toString;
                var arrayTypeStr = '[object Array]';
                var arr = [];
                var nodes = (mutable && self) || self.slice();
                var node;
                if (!self.length) {
                    return new (ArrayX.bind.apply(ArrayX, [void 0].concat(arr)))();
                }
                node = nodes.pop();
                do {
                    if (toString.call(node) === arrayTypeStr) {
                        nodes.push.apply(nodes, node);
                    }
                    else {
                        arr.push(node);
                    }
                } while (nodes.length && (node = nodes.pop()) !== undefined);
                arr.reverse(); // we reverse result to restore the original order
                return new (ArrayX.bind.apply(ArrayX, [void 0].concat(arr)))();
            };
            /**
             * Determines whether an array includes a certain element, returning true or false as appropriate.
             * @param searchElement The element to search for.
             * @param fromIndex The position in this array at which to begin searching for searchElement.
             */
            Arr["contains"] = function (obj, fromIndex) {
                return this.indexOf(obj, fromIndex) >= 0;
            };
            return Arr;
        }
        ArrayX.fromLength = function (size) {
            if (size === void 0) { size = 0; }
            return new (ArrayX.bind.apply(ArrayX, [void 0].concat(new Array(size))))();
        };
        ArrayX.fromRange = function (from, to) {
            var arr = [];
            var sgn = to >= from ? 1 : -1;
            for (var idx = from; idx != to; idx += sgn)
                arr.push(idx);
            return new (ArrayX.bind.apply(ArrayX, [void 0].concat(arr)))();
        };
        return ArrayX;
    }());
    ;
    return ArrayX;
})();
